from configcrunch import ReferencedDocumentNotFound
from configcrunch_tests.fixtures.documents import Base
from configcrunch_tests.acceptance.testcases import ConfigcrunchTestCase


class MergingCheckRefNotFound(ConfigcrunchTestCase):
    @classmethod
    def fixture_name(cls):
        return 'merging_check_ref_not_found'

    def test_invalid_ref(self):
        base = Base.from_yaml(self.fix_get_path('base.yml'))
        self.assertRaises(ReferencedDocumentNotFound, base.resolve_and_merge_references, [])

    def test_invalid_ref_relative(self):
        base = Base.from_yaml(self.fix_get_path('base_invalid_relative.yml'))
        self.assertRaises(ReferencedDocumentNotFound, base.resolve_and_merge_references, [])

    def test_invalid_ref_relative_parent(self):
        base = Base.from_yaml(self.fix_get_path('base_invalid_relative_parent.yml'))
        self.assertRaises(ReferencedDocumentNotFound, base.resolve_and_merge_references, [])

    def test_invalid_ref_relative_parent_two_layers(self):
        base = Base.from_yaml(self.fix_get_path('base_invalid_relative_parent_two_layers.yml'))
        self.assertRaises(ReferencedDocumentNotFound, base.resolve_and_merge_references, [self.fix_get_path('repo')])




from configcrunch import InvalidHeaderError, InvalidDocumentError
from configcrunch_tests.fixtures.documents import Base
from configcrunch_tests.acceptance.testcases import ConfigcrunchTestCase


class MergingCheckInvalidSubdoc(ConfigcrunchTestCase):
    @classmethod
    def fixture_name(cls):
        return 'merging_check_invalid_subdoc'

    def test_invalid(self):
        self.assertRaises(InvalidHeaderError, Base.from_yaml, self.fix_get_path('base_invalid.yml'))

    def test_invalid_in_ref(self):
        base = Base.from_yaml(self.fix_get_path('base_valid.yml'))
        self.assertRaises(InvalidHeaderError, base.resolve_and_merge_references, [self.fix_get_path('repo')])

    def test_invalid_empty(self):
        self.assertRaises(InvalidDocumentError, Base.from_yaml, self.fix_get_path('base_empty.yml'))

    def test_invalid_level_empty(self):
        base = Base.from_yaml(self.fix_get_path('base_valid_level_empty.yml'))
        self.assertRaises(InvalidDocumentError, base.resolve_and_merge_references, [self.fix_get_path('repo')])



import unittest

from configcrunch_tests.fixtures.documents import Base


class InternalTest(unittest.TestCase):
    def setUp(self) -> None:
        self.doc = Base.from_dict({
            "more": {
                "key_before": "value_before"
            }
        })

    def test_internal_get_before_freeze(self):
        self.assertEqual({"key_before": "value_before"}, self.doc.internal_get("more"))
        self.assertEqual(None, self.doc.internal_get("doesnt_exist"))

    def test_internal_set_before_freeze(self):
        self.doc.internal_set("more", "new_more")
        self.doc.internal_set("new_key", "new_new_key")
        self.assertEqual("new_more", self.doc.internal_get("more"))
        self.assertEqual("new_new_key", self.doc.internal_get("new_key"))

    def test_internal_contains_before_freeze(self):
        self.assertTrue(self.doc.internal_contains("more"))
        self.assertFalse(self.doc.internal_contains("doesnt_exist"))

    def test_internal_delete_before_freeze(self):
        self.doc.internal_delete("more")
        self.assertFalse(self.doc.internal_contains("more"))
        # Deleting must work without an error
        self.doc.internal_delete("doesnt_exist")

    def test_internal_access_before_freeze(self):
        with self.doc.internal_access():
            self.doc["more"] = "new_more"
            self.doc["new_key"] = "new_new_key"
            self.doc.internal_set("internal_new_key", "new_internal_new_key")
        self.assertEqual("new_more", self.doc.internal_get("more"))
        self.assertEqual("new_new_key", self.doc.internal_get("new_key"))
        self.assertEqual("new_internal_new_key", self.doc.internal_get("internal_new_key"))

    def test_internal_get_after_freeze(self):
        self.doc.freeze()
        self.assertEqual({"key_before": "value_before"}, self.doc.internal_get("more"))
        self.assertEqual(None, self.doc.internal_get("doesnt_exist"))

    def test_internal_set_after_freeze(self):
        self.doc.freeze()
        self.doc.internal_set("more", "new_more")
        self.doc.internal_set("new_key", "new_new_key")
        self.assertEqual("new_more", self.doc.internal_get("more"))
        self.assertEqual("new_new_key", self.doc.internal_get("new_key"))

    def test_internal_contains_after_freeze(self):
        self.doc.freeze()
        self.assertTrue(self.doc.internal_contains("more"))
        self.assertFalse(self.doc.internal_contains("doesnt_exist"))

    def test_internal_delete_after_freeze(self):
        self.doc.freeze()
        self.doc.internal_delete("more")
        self.assertFalse(self.doc.internal_contains("more"))
        # Deleting must work without an error
        self.doc.internal_delete("doesnt_exist")

    def test_internal_access_after_freeze(self):
        self.doc.freeze()
        with self.doc.internal_access():
            self.doc["more"] = "new_more"
            self.doc["new_key"] = "new_new_key"
            self.doc.internal_set("internal_new_key", "new_internal_new_key")
        self.assertEqual("new_more", self.doc.internal_get("more"))
        self.assertEqual("new_new_key", self.doc.internal_get("new_key"))
        self.assertEqual("new_internal_new_key", self.doc.internal_get("internal_new_key"))



from schema import SchemaError

from configcrunch_tests.acceptance.testcases import ConfigcrunchTestCase


class NegativeValidation(ConfigcrunchTestCase):
    @classmethod
    def fixture_name(cls):
        return 'negative_validation'

    def test_invalid_main(self):
        doc = self.load_base('negative.yml', [])
        self.assertRaises(SchemaError, doc.validate)

    def test_invalid_subdoc(self):
        doc = self.load_base('negative_subdocument.yml', [])
        self.assertRaises(SchemaError, doc.validate)



from configcrunch_tests.acceptance.testcases import ConfigcrunchTestCase


class MergingListDuplicates(ConfigcrunchTestCase):
    @classmethod
    def fixture_name(cls):
        return 'merging_list_duplicates'

    def test_lists(self):
        self.assertDocEqualMerging(
            'expected.yml',
            'base.yml',
            ['repo']
        )



from configcrunch._main import YamlConfigDocument, DocReference, load_multiple_yml, \
    ConfigcrunchError, ReferencedDocumentNotFound, CircularDependencyError, \
    VariableProcessingError, InvalidDocumentError, InvalidHeaderError, InvalidRemoveError

# Constants
REF = "$ref"
REMOVE = "$remove"
REMOVE_FROM_LIST_PREFIX = REMOVE + "::"


def variable_helper(func):
    orig_doc = ""
    if hasattr(func, "__doc__") and func.__doc__ is not None:
        orig_doc = func.__doc__
    func.__doc__ = """.. admonition:: Variable Helper

                  Can be used inside configuration files.

""" + orig_doc
    func.__is_variable_helper = True
    return func


try:
    import yaml
    def ycd_representer(dumper, data):
        return dumper.represent_mapping('!' + data.__class__.__name__, data.items())

    yaml.add_multi_representer(YamlConfigDocument, ycd_representer)
except ImportError:
    pass


# Public classes and functions
__all__ = [
    'YamlConfigDocument',
    'DocReference',
    'variable_helper',
    'load_multiple_yml',

    'ConfigcrunchError',
    'ReferencedDocumentNotFound',
    'CircularDependencyError',
    'VariableProcessingError',
    'InvalidDocumentError',
    'InvalidHeaderError',
    'InvalidRemoveError'
]



from configcrunch_tests.acceptance.testcases import ConfigcrunchTestCase


class MergingRelativeRefInRepo(ConfigcrunchTestCase):
    @classmethod
    def fixture_name(cls):
        return 'merging_relative_ref_in_repo'

    def test_one_repo(self):
        self.assertDocEqualMerging(
            'expected_one_repo.yml',
            'base.yml',
            ['repo1']
        )

    def test_parent_repo(self):
        self.assertDocEqualMerging(
            'expected_one_repo.yml',
            'base_parent_repo.yml',
            ['repo_parent_directory']
        )

    def test_two_repos(self):
        self.assertDocEqualMerging(
            'expected_two_repo.yml',
            'base.yml',
            ['repo1', 'repo2']
        )



from configcrunch_tests.acceptance.testcases import ConfigcrunchTestCase


class MergingRemove(ConfigcrunchTestCase):
    @classmethod
    def fixture_name(cls):
        return 'merging_remove'

    def test_dict_ref(self):
        self.assertDocEqualMerging(
            'expected.yml',
            'base.yml',
            ['repo']
        )

    def test_remove_subdoc(self):
        self.assertDocEqualMerging(
            'expected_subdoc.yml',
            'base_subdoc.yml',
            ['repo']
        )

    def test_remove_subdoc_multi_repo(self):
        self.assertDocEqualMerging(
            'expected_subdoc_multi_repo.yml',
            'base_subdoc_multi_repo.yml',
            ['repo', 'repo2']
        )

    def test_multi_level_list(self):
        self.assertDocEqualMerging(
            'multi_list_merge_expected.yml',
            'multi_list_merge_base.yml',
            ['repo']
        )



from configcrunch_tests.acceptance.testcases import ConfigcrunchTestCase


class MergingEasyTest(ConfigcrunchTestCase):
    @classmethod
    def fixture_name(cls):
        return 'merging_easy'

    def test_same(self):
        self.assertDocEqualMerging(
            'easy_expected.yml',
            'easy.yml',
            []
        )

    def test_different(self):
        doc = self.load_base('something_else.yml', [])
        expected_result = self.fix_get_yml('easy.yml')

        self.assertNotEqual(expected_result, doc.to_dict())
        self.assertValidDoc(doc)

    def test_different_direct(self):
        easy_doc = self.load_base('easy.yml', [])
        other_doc = self.load_base('something_else.yml', [])

        self.assertNotEqual(easy_doc.to_dict(), other_doc.to_dict())
        self.assertValidDoc(easy_doc)
        self.assertValidDoc(other_doc)



import unittest
from configcrunch._main import _test__subdoc_specs

FIXTURE = {
    "level1": {
        "level2": {
            "list2": ["hello", "world", "how", "are", "you"],
            "dict2": {"lev2_1": "lev2_1_val", "lev2_2": "lev2_2_val"}
        },
        "list1": ["hi", "level 1"],
        "dict1": {
            "lev1_1": "lev1_1_val",
            "lev1_2": "lev1_2_val",
            "lev1_3": "lev1_3_val"
        }
    },
    "direct_dict": {"replace": "me"},
    "direct": "hi",
    "list": ["list1", "list2"]
}


class SubdocSpecsTest(unittest.TestCase):
    def test_subdoc_specs(self):
        self.assertEqual(
            ({
                "level1": {
                    "level2": {
                        "list2": ["hello", "world", "how", "are", "you"],
                        "dict2": {"lev2_1": "lev2_1_val", "lev2_2": "lev2_2_val"}
                    },
                    "list1": ["hi", "level 1"],
                    "dict1": {
                        "lev1_1": "lev1_1_val",
                        "lev1_2": "lev1_2_val",
                        "lev1_3": "lev1_3_val"
                    }
                },
                "direct_dict": "REPLACED",
                "direct": "hi",
                "list": ["list1", "list2"]
            }, type),
            _test__subdoc_specs("direct_dict", type, FIXTURE, "REPLACED")
        )
        self.assertEqual(
            ({
                "level1": {
                    "level2": {
                        "list2": ["hello", "world", "how", "are", "you"],
                        "dict2": {"lev2_1": "lev2_1_val", "lev2_2": "lev2_2_val"}
                    },
                    "list1": ["hi", "level 1"],
                    "dict1": {
                        "lev1_1": "lev1_1_val",
                        "lev1_2": "lev1_2_val",
                        "lev1_3": "lev1_3_val"
                    }
                },
                "direct_dict": {"replace": "me"},
                "direct": "hi",
                "list": "REPLACED"
            }, type),
            _test__subdoc_specs("list", type, FIXTURE, "REPLACED")
        )
        self.assertEqual(
            ({
                "level1": {
                    "level2": {
                        "list2": ["hello", "world", "how", "are", "you"],
                        "dict2": {"lev2_1": "lev2_1_val", "lev2_2": "lev2_2_val"}
                    },
                    "list1": ["hi", "level 1"],
                    "dict1": {
                        "lev1_1": "lev1_1_val",
                        "lev1_2": "lev1_2_val",
                        "lev1_3": "lev1_3_val"
                    }
                },
                "direct_dict": {"replace": "me"},
                "direct": "hi",
                "list": ["REPLACED", "REPLACED"]
            }, type),
            _test__subdoc_specs("list[]", type, FIXTURE, "REPLACED")
        )
        self.assertEqual(
            ({
                "level1": {
                    "level2": {
                        "list2": ["hello", "world", "how", "are", "you"],
                        "dict2": {"lev2_1": "lev2_1_val", "lev2_2": "lev2_2_val"}
                    },
                    "list1": ["hi", "level 1"],
                    "dict1": {
                        "lev1_1": "lev1_1_val",
                        "lev1_2": "lev1_2_val",
                        "lev1_3": "lev1_3_val"
                    }
                },
                "direct_dict": {"replace": "REPLACED"},
                "direct": "hi",
                "list": ["list1", "list2"]
            }, type),
            _test__subdoc_specs("direct_dict[]", type, FIXTURE, "REPLACED")
        )
        self.assertEqual(
            ({
                "level1": {
                    "level2": {
                        "list2": ["hello", "world", "how", "are", "you"],
                        "dict2": {"lev2_1": "lev2_1_val", "lev2_2": "lev2_2_val"}
                    },
                    "list1": ["REPLACED", "REPLACED"],
                    "dict1": {
                        "lev1_1": "lev1_1_val",
                        "lev1_2": "lev1_2_val",
                        "lev1_3": "lev1_3_val"
                    }
                },
                "direct_dict": {"replace": "me"},
                "direct": "hi",
                "list": ["list1", "list2"]
            }, type),
            _test__subdoc_specs("level1/list1[]", type, FIXTURE, "REPLACED")
        )
        self.assertEqual(
            ({
                "level1": {
                    "level2": {
                        "list2": ["hello", "world", "how", "are", "you"],
                        "dict2": "REPLACED"
                    },
                    "list1": ["hi", "level 1"],
                    "dict1": {
                        "lev1_1": "lev1_1_val",
                        "lev1_2": "lev1_2_val",
                        "lev1_3": "lev1_3_val"
                    }
                },
                "direct_dict": {"replace": "me"},
                "direct": "hi",
                "list": ["list1", "list2"]
            }, type),
            _test__subdoc_specs("level1/level2/dict2", type, FIXTURE, "REPLACED")
        )
        self.assertEqual(
            ({
                "level1": {
                    "level2": {
                        "list2": ["hello", "world", "how", "are", "you"],
                        "dict2": {"lev2_1": "REPLACED", "lev2_2": "REPLACED"}
                    },
                    "list1": ["hi", "level 1"],
                    "dict1": {
                        "lev1_1": "lev1_1_val",
                        "lev1_2": "lev1_2_val",
                        "lev1_3": "lev1_3_val"
                    }
                },
                "direct_dict": {"replace": "me"},
                "direct": "hi",
                "list": ["list1", "list2"]
            }, type),
            _test__subdoc_specs("level1/level2/dict2[]", type, FIXTURE, "REPLACED")
        )



# See the Rust source code for docstrings, query them via Python or see the documentation..
from __future__ import annotations

from abc import abstractmethod
from typing import List, Callable, Type, Optional, Union, TypeVar, final, Tuple, Any, ContextManager, Self

from schema import Schema

from configcrunch import variable_helper


T = TypeVar('T', bound=YamlConfigDocument)


class ConfigcrunchError(Exception): ...
class ReferencedDocumentNotFound(ConfigcrunchError): ...
class CircularDependencyError(ConfigcrunchError): ...
class VariableProcessingError(ConfigcrunchError): ...
class InvalidDocumentError(ConfigcrunchError): ...
class InvalidHeaderError(InvalidDocumentError):...
class InvalidRemoveError(InvalidDocumentError): ...


def load_multiple_yml(doc_type: Type[T], *in_args: str) -> T: ...


class YamlConfigDocument:
    path: Optional[str]
    parent_doc: Optional[YamlConfigDocument]
    absolute_paths: List[str]
    
    def __init__(
            self, document: Self|dict, path: Optional[str] = None, parent_doc: Optional[YamlConfigDocument] = None
    ):
        """Manual constructor. It's recommended to use from_yaml or from_dict instead."""
        ...

    @property
    def doc(self) -> dict:
        """READONLY representation of the document, this is also accessible via __getitem__ etc. ONLY after calling freeze()"""
        ...

    @classmethod
    @final
    def from_yaml(cls, path_to_yaml: str) -> Self: ...
    @classmethod
    @final
    def from_dict(cls, dict: dict) -> Self: ...
    @final
    def freeze(self): ...
    @classmethod
    @abstractmethod
    def header(cls) -> str: ...
    @classmethod
    @abstractmethod
    def schema(cls) -> Schema: ...
    @classmethod
    @abstractmethod
    def subdocuments(cls) -> List[Tuple[str, Type[YamlConfigDocument]]]: ...
    def validate(self) -> bool: ...
    @final
    def resolve_and_merge_references(self, lookup_paths: List[str]) -> Self: ...
    @final
    def process_vars(self) -> Self: ...
    def process_vars_for(self, target: str, additional_helpers: List[Callable]|None = None) -> str: ...
    @variable_helper
    def parent(self) -> Optional[YamlConfigDocument]: ...
    def __repr__(self) -> str: ...
    def __str__(self): ...
    def error_str(self) -> str: ...
    def __len__(self): ...
    def __getitem__(self, key): ...
    def __setitem__(self, key, value): ...
    def __delitem__(self, key): ...
    def __iter__(self): ...
    def items(self): ...
    def to_dict(self): ...
    def internal_get(self, key: str) -> Any: ...
    def internal_set(self, key: str, val: Any): ...
    def internal_contains(self, key: str) -> bool: ...
    def internal_delete(self, key: str): ...
    def internal_access(self) -> ContextManager: ...
    # These DO NOT ACTUALLY EXIST on the parent object.
    def _initialize_data_before_merge(self, data: dict) -> dict:
        """
        May be used to initialize the document by adding / changing data.

        Called before doing anything else in resolve_and_merge_references.
        Use this for setting default values.
        The internal data is passed as an argument and can be mutated.
        The changed data MUST be returned again.
        """
        ...
    def _initialize_data_after_merge(self, data: dict) -> dict:
        """
        May be used to initialize the document by adding / changing data.

        Called after resolve_and_merge_references.
        Use this for setting default values.
        The internal data is passed as an argument and can be mutated.
        The changed data MUST be returned again.
        """
        ...
    def _initialize_data_after_variables(self, data: dict) -> dict:
        """
        May be used to initialize the document by adding / changing data.

        Called after process_vars.
        Use this for setting internal values based on processed values in the document.
        The internal data is passed as an argument and can be mutated.
        The changed data MUST be returned again.
        """
        ...
    def _initialize_data_after_freeze(self):
        """
        May be used to initialize the document by adding / changing data.

        Called after freeze.
        Use this for setting internal values based on processed values in the document.
        You can access the data using the self.doc property or by getting it from self (self[...]).
        """
        ...


class DocReference:
    referenced_type: Type[YamlConfigDocument]

    def __init__(self, referenced_doc_type: Type[YamlConfigDocument]): ...
    def validate(self, data): ...


def _test__subdoc_specs(path: str, type: Type[Any], input: dict) -> Tuple[dict, str, Any, bool, Type[Any]]: ...



from configcrunch_tests.acceptance.testcases import ConfigcrunchTestCase


class MergingThreeRepos(ConfigcrunchTestCase):
    @classmethod
    def fixture_name(cls):
        return 'merging_three_repos'

    def test_no_subdoc(self):
        self.assertDocEqualMerging(
            'expected/no_subdoc.yml',
            'no_subdoc.yml',
            ['repo1', 'repo2', 'repo3']
        )

    def test_one_layer_subdoc(self):
        self.assertDocEqualMerging(
            'expected/one_level.yml',
            'one_level.yml',
            ['repo1', 'repo2', 'repo3']
        )

    def test_two_layers_subdoc(self):
        self.assertDocEqualMerging(
            'expected/two_level.yml',
            'two_level.yml',
            ['repo1', 'repo2', 'repo3']
        )



from configcrunch import YamlConfigDocument


# Called from Rust code (YamlConfigDocument::to_dict).
def frozen_ycd_to_dict(ycd: YamlConfigDocument):
    return {ycd.header(): frozen_ycd_to_dict_rec(ycd)}


def frozen_ycd_to_dict_rec(input):
    """ Recursively removes all YamlConfigDocuments and replaces them by their doc dictionary."""
    if isinstance(input, YamlConfigDocument):
        return frozen_ycd_to_dict_rec(input.doc.copy())
    elif isinstance(input, dict):
        new_dict = input.copy()
        for key, val in new_dict.items():
            new_dict[key] = frozen_ycd_to_dict_rec(val)
        return new_dict
    elif isinstance(input, list):
        new_list = []
        for item in input.copy():
            new_list.append(frozen_ycd_to_dict_rec(item))
        return new_list
    return input



from configcrunch import CircularDependencyError
from configcrunch_tests.fixtures.documents import Base
from configcrunch_tests.acceptance.testcases import ConfigcrunchTestCase


class MergingCheckRefNotFound(ConfigcrunchTestCase):
    @classmethod
    def fixture_name(cls):
        return 'merging_check_infinite_recursion'

    def test_in_itself(self):
        base = Base.from_yaml(self.fix_get_path('in_itself.yml'))
        self.assertRaises(CircularDependencyError, base.resolve_and_merge_references, [self.fix_get_path('repo')])

    def test_via_child(self):
        base = Base.from_yaml(self.fix_get_path('via_child.yml'))
        self.assertRaises(CircularDependencyError, base.resolve_and_merge_references, [self.fix_get_path('repo')])



from configcrunch import load_multiple_yml
from configcrunch_tests.fixtures.documents import Base
from configcrunch_tests.acceptance.testcases import ConfigcrunchTestCase, deep_sort


class AdvancedLoader(ConfigcrunchTestCase):
    @classmethod
    def fixture_name(cls):
        return 'advanced_loader'

    def test_none(self):
        doc = load_multiple_yml(
            Base, self.fix_get_path('deep.yml'), self.fix_get_path('middle.yml'), self.fix_get_path('top.yml')
        )
        expected_result = self.fix_get_yml('expected.yml')

        self.assertDictEqual(deep_sort(expected_result), deep_sort(doc.to_dict()))
        self.assertValidDoc(doc)



import unittest
from copy import copy

from configcrunch_tests.fixtures.documents import Base
FIX = {"more": "data_modified_after_merge"}


class AfterInitTest(unittest.TestCase):
    def setUp(self) -> None:
        self.doc = Base.from_dict({
            "more": {
                "key_before": "value_before"
            }
        })

    def test_initialize_data_before_merge(self):
        def _initialize_data_before_merge(data):
            return copy(FIX)

        setattr(self.doc, '_initialize_data_before_merge', _initialize_data_before_merge)
        self.doc.resolve_and_merge_references([])
        self.assertEqual({"base": copy(FIX)}, self.doc.to_dict())

    def test_initialize_data_after_merge(self):
        def _initialize_data_after_merge(data):
            return copy(FIX)

        setattr(self.doc, '_initialize_data_after_merge', _initialize_data_after_merge)
        self.doc.resolve_and_merge_references([])
        self.assertEqual({"base": copy(FIX)}, self.doc.to_dict())

    def test_initialize_data_after_variables(self):
        def _initialize_data_after_variables(data):
            return copy(FIX)

        setattr(self.doc, '_initialize_data_after_variables', _initialize_data_after_variables)
        self.doc.process_vars()
        self.assertEqual({"base": copy(FIX)}, self.doc.to_dict())

    def test_initialize_data_after_freeze(self):
        def _initialize_data_after_freeze():
            self.doc["more"] = copy(FIX)

        setattr(self.doc, '_initialize_data_after_freeze', _initialize_data_after_freeze)
        self.doc.freeze()
        self.assertEqual({"more": copy(FIX)}, self.doc.items())



from configcrunch_tests.acceptance.testcases import ConfigcrunchTestCase


class MergingSubdoc(ConfigcrunchTestCase):
    @classmethod
    def fixture_name(cls):
        return 'merging_subdoc'

    def test_dict_ref(self):
        self.assertDocEqualMerging(
            'expected/dict_ref.yml',
            'dict_ref.yml',
            ['repo']
        )

    def test_dict_ref_with_maindoc_ref(self):
        self.assertDocEqualMerging(
            'expected/dict_ref_with_maindoc_ref.yml',
            'dict_ref_with_maindoc_ref.yml',
            ['repo']
        )

    def test_direct_ref(self):
        self.assertDocEqualMerging(
            'expected/direct_ref.yml',
            'direct_ref.yml',
            ['repo']
        )

    def test_direct_ref_with_maindoc_ref(self):
        self.assertDocEqualMerging(
            'expected/direct_ref_with_maindoc_ref.yml',
            'direct_ref_with_maindoc_ref.yml',
            ['repo']
        )

    def test_list_ref(self):
        self.assertDocEqualMerging(
            'expected/list_ref.yml',
            'list_ref.yml',
            ['repo']
        )

    def test_list_ref_with_maindoc_ref(self):
        self.assertDocEqualMerging(
            'expected/list_ref_with_maindoc_ref.yml',
            'list_ref_with_maindoc_ref.yml',
            ['repo']
        )



import os
import unittest
from abc import abstractmethod, ABC
from collections.abc import Hashable

import yaml
from schema import SchemaError
from configcrunch_tests.fixtures.documents import Base


class ConfigcrunchTestCase(unittest.TestCase, ABC):

    def setUp(self):
        self.maxDiff = None

    @classmethod
    @abstractmethod
    def fixture_name(cls) -> str:
        pass

    @classmethod
    def __fixture_path(cls):
        return os.path.abspath(os.path.join(os.path.dirname(__file__), os.pardir, 'fixtures', cls.fixture_name()))

    @classmethod
    def fix_get_path(cls, path):
        return os.path.join(cls.__fixture_path(), path)

    @classmethod
    def fix_get_yml(cls, path):
        with open(cls.fix_get_path(path), 'r') as stream:
            return yaml.safe_load(stream)

    def load_base(self, path, lookup_paths):
        base = Base.from_yaml(self.fix_get_path(path))
        abs_lookup_paths = []
        for path in lookup_paths:
            abs_lookup_paths.append(self.fix_get_path(path))
        base.resolve_and_merge_references(abs_lookup_paths)
        return base

    def assertDocEqualMerging(self, expected_yml_file, input_yml_file, repo_folder_names):
        doc = self.load_base(input_yml_file, repo_folder_names)
        expected_result = self.fix_get_yml(expected_yml_file)

        self.assertDictEqual(deep_sort(expected_result), deep_sort(doc.to_dict()))
        self.assertValidDoc(doc)

    def assertDocEqualVariables(self, expected_yml_file, input_yml_file):
        doc = self.load_base(input_yml_file, [])
        doc.process_vars()
        expected_result = self.fix_get_yml(expected_yml_file)

        self.assertDictEqual(deep_sort(expected_result), deep_sort(doc.to_dict()))
        self.assertValidDoc(doc)

    def assertValidDoc(self, doc):
        try:
            self.assertTrue(doc.validate())
        except SchemaError:
            self.fail("The document is not valid")


""" Helper functions to guarantee list orders follow"""


def sortby(x):
    """Return integer hash for all non-numbers (if possible, otherwise inf). To make everything sortable"""
    if x.__class__.__lt__ != x.__lt__:
        if isinstance(x, Hashable):
            return hash(x)
        return float('inf')
    return x


def deep_sort(obj):
    """Recursively sort list or nested lists"""
    if isinstance(obj, dict):
        _sorted = {}
        for key in sorted(obj, key=sortby):
            _sorted[key] = deep_sort(obj[key])

    elif isinstance(obj, list):
        new_list = []
        for val in obj:
            new_list.append(deep_sort(val))
        _sorted = sorted(new_list, key=sortby)

    else:
        _sorted = obj

    return _sorted



from configcrunch_tests.acceptance.testcases import ConfigcrunchTestCase


class Variables(ConfigcrunchTestCase):
    @classmethod
    def fixture_name(cls):
        return 'variables'

    def test_none(self):
        self.assertDocEqualVariables(
            'none.yml',
            'none.yml'
        )

    def test_one_level(self):
        self.assertDocEqualVariables(
            'expected/one_level.yml',
            'one_level.yml'
        )

    def test_helpers(self):
        self.assertDocEqualVariables(
            'expected/helper_calls.yml',
            'helper_calls.yml'
        )

    def test_accessing_child_vars(self):
        self.assertDocEqualVariables(
            'expected/accessing_child_vars.yml',
            'accessing_child_vars.yml'
        )

    def test_complex(self):
        self.assertDocEqualVariables(
            'expected/complex.yml',
            'complex.yml'
        )

    def test_not_working(self):
        """
        When calling parent() and accessing a field on the parent with variables in it,
        the variables are not correctly resolved*. This also happens when trying to access a field of a
        subdocument of the parent that contains a variable and hasn't been processed yet.

        Therefor such things are not currently not allowed. If this becomes possible in the future
        feel free to rename and change this test.

        *: To be exact: The variable value is copied over to the subdocument as-is and processed there in it's context.
        """
        self.skipTest("Order of internal dict no longer guaranteed, so the result varies.")
        return
        self.assertDocEqualVariables(
            'expected/not_working.yml',
            'not_working.yml'
        )



from configcrunch_tests.acceptance.testcases import ConfigcrunchTestCase


class MergingOneRef(ConfigcrunchTestCase):
    @classmethod
    def fixture_name(cls):
        return 'merging_one_ref'

    def test_one_repo_nothing_to_merge(self):
        self.assertDocEqualMerging(
            'expected.yml',
            'base.yml',
            ['repo1']
        )

    def test_one_repo_something_to_merge(self):
        self.assertDocEqualMerging(
            'expected_with_something_to_merge.yml',
            'base_with_something_to_merge.yml',
            ['repo1']
        )

    def test_two_repos_something_to_merge(self):
        self.assertDocEqualMerging(
            'expected_with_something_to_merge_two_repos.yml',
            'base_with_something_to_merge.yml',
            ['repo1', 'repo2']
        )



"""
Classes that implement YamlConfigDocument and represent
YAML documents for the tests.
"""
from typing import List, Tuple, Type

from schema import Schema, Optional

from configcrunch import YamlConfigDocument, DocReference, REMOVE, variable_helper


class Base(YamlConfigDocument):
    """
    Base test document. Example:

    base:
        str_field: string
        int_field: 12
        level_dict:
            xyz: !Level
        level_array:
            - !Level
        level_direct: !Level
        more: any

    All fields are optional.
    """
    @classmethod
    def header(cls) -> str:
        return "base"

    @classmethod
    def schema(cls) -> Schema:
        return Schema(
            {
                Optional('$ref'): str,  # reference to other Base documents
                Optional('str_field'): str,
                Optional('int_field'): int,
                Optional('level_dict'): {
                    str: DocReference(Level)
                },
                Optional('level_array'): [DocReference(Level)],
                Optional('level_direct'): DocReference(Level),
                Optional('more'): lambda any: True
            }
        )

    @classmethod
    def subdocuments(cls) -> List[Tuple[str, Type[YamlConfigDocument]]]:
        return [
            ("level_dict[]", Level),
            ("level_array[]", Level),
            ("level_direct", Level),
        ]

    @variable_helper
    def simple_helper(self):
        return "simple"


class Level(YamlConfigDocument):
    """
    Level test document. Example:

    level:
        name: str
        base_ref: !Base
        more: any

    All fields (except for name) are optional.
    """

    @classmethod
    def header(cls) -> str:
        return "level"

    @classmethod
    def schema(cls) -> Schema:
        return Schema(
            {
                Optional('$ref'): str,  # reference to other Level documents
                Optional('$name'): str,  # key if in a dict
                'name': str,
                Optional('base_ref'): DocReference(Base),
                Optional('more'): lambda any: True
            }
        )

    @classmethod
    def subdocuments(cls) -> List[Tuple[str, Type[YamlConfigDocument]]]:
        return [
            ("base_ref", Base),
        ]

    @variable_helper
    def level_helper(self):
        return "level"

    @variable_helper
    def level_helper_taking_param(self, param: str):
        return f"level_param: {param}"
